'use strict';

var _ = require('lodash');
var slugify = require('slugify');
var capitalize = require('capitalize')

slugify.extend({ '+': 'plus' });
slugify.extend({ '-': 'minus' });
/**
 * Recursively converts a swagger type description into a typescript type, i.e., a model for our mustache
 * template.
 *
 * Not all type are currently supported, but they should be straightforward to add.
 *
 * @param swaggerType a swagger type definition, i.e., the right hand side of a swagger type definition.
 * @returns a recursive structure representing the type, which can be used as a template model.
 */
function convertType(data, [paramName, namespace], swaggerType, swagger) {
    var typespec = { description: swaggerType.description, isEnum: false, isNullable: !swaggerType.required, cardinality: swaggerType.required ? '' : '?' };

    if (swaggerType.hasOwnProperty('schema')) {
        return convertType(data, [paramName, namespace], swaggerType.schema);
    } else if (_.isString(swaggerType.$ref)) {
        typespec.tsType = 'ref';
        typespec.target = swaggerType.$ref.substring(swaggerType.$ref.lastIndexOf('/') + 1);
    } else if (swaggerType.hasOwnProperty('enum')) {
        const enumValues = _.uniqBy(swaggerType.enum.map(function (str) {
            let slug = slugify(str, { replacement: '_' });
            if (/^[\d]/.test(slug)) {
                slug = 'Int' + slug;
            }
            return {
                slug: slug,
                value: JSON.stringify(str)
            };
        }), 'slug');

        data.enums.push({
            name: paramName,
            namespace,
            values: enumValues
        });

        typespec.tsType = namespace + '.' + paramName;
        typespec.isAtomic = true;
        typespec.isEnum = true;

    } else if (swaggerType.type === 'string') {
        typespec.tsType = 'string';
    } else if (swaggerType.type === 'number' || swaggerType.type === 'integer') {
        typespec.tsType = 'number';
    } else if (swaggerType.type === 'boolean') {
        typespec.tsType = 'boolean';
    } else if (swaggerType.type === 'array') {
        typespec.elementType = convertType(data, [paramName, namespace], swaggerType.items);
        typespec.tsType = `Array<${typespec.elementType.target || typespec.elementType.tsType || 'any'}>`;
    } else /*if (swaggerType.type === 'object')*/ { //remaining types are created as objects
        if (swaggerType.minItems >= 0 && swaggerType.hasOwnProperty('title') && !swaggerType.$ref) {
            typespec.tsType = 'any';
        }
        else {
            typespec.tsType = 'object';
            typespec.properties = [];
            if (swaggerType.allOf) {
                _.forEach(swaggerType.allOf, function (ref) {
                    if(ref.$ref) {
                        let refSegments = ref.$ref.split('/');
                        let name = refSegments[refSegments.length - 1];
                        _.forEach(swagger.definitions, function (definition, definitionName) {
                            if (definitionName === name) {
                                var property = convertType(data, [definitionName, namespace + capitalize(paramName)], definition, swagger);
                                typespec.properties.push(...property.properties);
                            }
                        });
                    } else {
                        var property = convertType(data, [paramName, namespace], ref);
                        typespec.properties.push(...property.properties);
                    }
                });
            }

            _.forEach(swaggerType.properties, function (propertyType, propertyName) {
                var property = convertType(data, [propertyName, namespace + capitalize(paramName)], propertyType);
                property.name = propertyName;
                typespec.properties.push(property);
            });

            if (swaggerType.required) {
                _.forEach(typespec.properties, function(property) {
                    if (swaggerType.required.indexOf(property.name) !== -1) {
                        property.cardinality = '';
                        property.isNullable = false;
                    }
                });
            }
        }
    }

    // Since Mustache does not provide equality checks, we need to do the case distinction via explicit booleans
    typespec.isRef = typespec.tsType === 'ref';
    typespec.isObject = typespec.tsType === 'object';
    typespec.isArray = typespec.tsType.indexOf("Array<") > -1;
    typespec.isAtomic = typespec.isAtomic || _.includes(['string', 'number', 'boolean', 'any'], typespec.tsType);

    return typespec;

}

module.exports.convertType = convertType;

